package terminal;

import java.io.IOException;

import com.googlecode.lanterna.TerminalPosition;
import com.googlecode.lanterna.input.*;
import com.googlecode.lanterna.terminal.*;

public class LispTerminal implements Runnable {

    private IOSafeTerminal terminal;
    private boolean isFinished;
    private String currentLine;
    private TerminalPosition origin;

    public LispTerminal() {
        try {
            Terminal unsafe = new DefaultTerminalFactory().createTerminal();
            this.terminal = IOSafeTerminalAdapter.createRuntimeExceptionConvertingAdapter(unsafe);
        } catch (IOException e) {
            // TODO
            e.printStackTrace();
        }

        this.isFinished = false;
        this.currentLine = "";
        this.origin = terminal.getCursorPosition();
    }

    public void run() {
        while (!isFinished) {
            handleKeyStroke(getKeyStroke());
            terminal.flush();
        }

        terminal.close();
    }

    private KeyStroke getKeyStroke() {
        KeyStroke keyStroke = null;

        try {
            keyStroke = terminal.readInput();
        } catch (IllegalStateException e) {
            // TODO - lantera is trying to exit private mode on ctrl-c (which we didn't enter)
            terminal.putCharacter('`');
            isFinished = true;
        }

        return keyStroke;
    }

    private void handleKeyStroke(KeyStroke keyStroke) {
        if (keyStroke == null)
            return;

        KeyType keyType = keyStroke.getKeyType();

        if (keyType == KeyType.ArrowLeft)
            moveCursorLeft();
        else if (keyType == KeyType.ArrowRight)
            moveCursorRight();
        else if (keyType == KeyType.Enter)
            doEnter();
        else if (keyType == KeyType.Backspace)
            doBackspace();
        else if (keyType == KeyType.Delete)
            doDelete();
        else if (keyType == KeyType.Character)
            doCharacter(keyStroke);
    }

    private void doEnter() {
        terminal.putCharacter('\n');
        currentLine = "";
        origin = terminal.getCursorPosition();
    }

    private void moveCursorLeft() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition))
            if (cursorPosition.getColumn() == 0)
                terminal.setCursorPosition(terminal.getTerminalSize().getColumns(), cursorPosition.getRow() - 1);
            else
                terminal.setCursorPosition(cursorPosition.withRelativeColumn(-1));
    }

    private boolean isPossibleToMoveLeft(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) > 0;
    }

    private int distanceFromOrigin(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn()
               + (terminal.getTerminalSize().getColumns() * (cursorPosition.getRow() - origin.getRow()));
    }

    private void moveCursorRight() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition))
            advanceCursor(cursorPosition);
    }

    private void advanceCursor(TerminalPosition cursorPosition) {
        if (isCursorAtEndOfRow(cursorPosition))
            terminal.setCursorPosition(0, cursorPosition.getRow() + 1);
        else
            terminal.setCursorPosition(cursorPosition.withRelativeColumn(1));
    }

    private boolean isCursorAtEndOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() == terminal.getTerminalSize().getColumns() - 1;
    }

    private boolean isPossibleToMoveRight(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) < currentLine.length();
    }

    private void doBackspace() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition)) {
            String remaining = currentLine.substring(distanceFromOrigin(cursorPosition), currentLine.length());
            currentLine = currentLine.substring(0, distanceFromOrigin(cursorPosition) - 1) + remaining;

            if (cursorPosition.getColumn() == 0)
                terminal.setCursorPosition(terminal.getTerminalSize().getColumns(), cursorPosition.getRow() - 1);
            else
                terminal.setCursorPosition(cursorPosition.withRelativeColumn(-1));

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');

            if (cursorPosition.getColumn() == 0)
                terminal.setCursorPosition(terminal.getTerminalSize().getColumns(), cursorPosition.getRow() - 1);
            else
                terminal.setCursorPosition(cursorPosition.withRelativeColumn(-1));
        }
    }

    private void doDelete() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = currentLine.substring(distanceFromOrigin(cursorPosition) + 1, currentLine.length());
            currentLine = currentLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');
            terminal.setCursorPosition(cursorPosition);
        }
    }

    private void doCharacter(KeyStroke keyStroke) {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = keyStroke.getCharacter()
                               + currentLine.substring(distanceFromOrigin(cursorPosition), currentLine.length());
            currentLine = currentLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            advanceCursor(cursorPosition);
        } else {
            terminal.putCharacter(keyStroke.getCharacter());
            currentLine += keyStroke.getCharacter();

            advanceCursor(cursorPosition);
        }
    }

}
